/* Added code to display tables without horizontal scrollbars */
.wy-table-responsive table td, .wy-table-responsive table th {
    white-space: normal;
}

/* Added code to override the default content width in ReadtheDocs */
.wy-nav-content {
    max-width: none;
}
